@@ -20,6 +20,6 @@ static inline void core_get_tls(CoreEntry *pcore, u32 *ptls)
 }
 
 
-int sigreturn_prep_fpu_frame(struct thread_restore_args *args, CoreEntry *core);
+int restore_fpu(struct rt_sigframe *sigframe, CoreEntry *core);
 
 #endif
