@@ -71,11 +71,13 @@ static void pr_info_ipc_sem_entry(const IpcSemEntry *sem)
 
 static int dump_ipc_sem_set(int fd, const IpcSemEntry *sem)
 {
+	size_t rounded;
 	int ret, size;
 	u16 *values;
 
 	size = sizeof(u16) * sem->nsems;
-	values = xmalloc(size);
+	rounded = round_up(size, sizeof(u64));
+	values = xmalloc(rounded);
 	if (values == NULL) {
 		pr_err("Failed to allocate memory for semaphore set values\n");
 		ret = -ENOMEM;
@@ -89,7 +91,8 @@ static int dump_ipc_sem_set(int fd, const IpcSemEntry *sem)
 	}
 	pr_info_ipc_sem_array(sem->nsems, values);
 
-	ret = write_img_buf(fd, values, round_up(size, sizeof(u64)));
+	memzero((void *)values + size, rounded - size);
+	ret = write_img_buf(fd, values, rounded);
 	if (ret < 0) {
 		pr_err("Failed to write IPC message data\n");
 		goto out;
@@ -185,7 +188,7 @@ static int dump_ipc_msg_queue_messages(int fd, const IpcMsgEntry *msq,
 	}
 
 	msgmax += sizeof(struct msgbuf);
-	message = xmalloc(msgmax);
+	message = xmalloc(round_up(msgmax, sizeof(u64)));
 	if (message == NULL) {
 		pr_err("Failed to allocate memory for IPC message\n");
 		return -ENOMEM;
@@ -193,6 +196,7 @@ static int dump_ipc_msg_queue_messages(int fd, const IpcMsgEntry *msq,
 
 	for (msg_cnt = 0; msg_cnt < msg_nr; msg_cnt++) {
 		IpcMsg msg = IPC_MSG__INIT;
+		size_t rounded;
 
 		ret = msgrcv(msq->desc->id, message, msgmax, msg_cnt, IPC_NOWAIT | MSG_COPY);
 		if (ret < 0) {
@@ -210,7 +214,10 @@ static int dump_ipc_msg_queue_messages(int fd, const IpcMsgEntry *msq,
 			pr_err("Failed to write IPC message header\n");
 			break;
 		}
-		ret = write_img_buf(fd, message->mtext, round_up(msg.msize, sizeof(u64)));
+
+		rounded = round_up(msg.msize, sizeof(u64));
+		memzero(((void *)message->mtext + msg.msize), rounded - msg.msize);
+		ret = write_img_buf(fd, message->mtext, rounded);
 		if (ret < 0) {
 			pr_err("Failed to write IPC message data\n");
 			break;
@@ -448,20 +455,28 @@ static int dump_ipc_data(const struct cr_fdset *fdset)
 	return 0;
 }
 
-int dump_ipc_ns(int ns_pid, const struct cr_fdset *fdset)
+int dump_ipc_ns(int ns_pid, int ns_id)
 {
 	int ret;
+	struct cr_fdset *fdset;
+
+	fdset = cr_fdset_open(ns_id, IPCNS, O_DUMP);
+	if (fdset == NULL)
+		return -1;
 
 	ret = switch_ns(ns_pid, &ipc_ns_desc, NULL);
 	if (ret < 0)
-		return ret;
+		goto err;
 
 	ret = dump_ipc_data(fdset);
 	if (ret < 0) {
 		pr_err("Failed to write IPC namespace data\n");
-		return ret;
+		goto err;
 	}
-	return 0;
+
+err:
+	close_cr_fdset(&fdset);
+	return ret < 0 ? -1 : 0;
 }
 
 void ipc_sem_handler(int fd, void *obj)
@@ -471,11 +486,11 @@ void ipc_sem_handler(int fd, void *obj)
 	int size;
 
 	pr_msg("\n");
-	size = sizeof(u16) * e->nsems;
+	size = round_up(sizeof(u16) * e->nsems, sizeof(u64));
 	values = xmalloc(size);
 	if (values == NULL)
 		return;
-	if (read_img_buf(fd, values, round_up(size, sizeof(u64))) <= 0) {
+	if (read_img_buf(fd, values, size) <= 0) {
 		xfree(values);
 		return;
 	}
