@@ -1,3 +1,5 @@
+#define _XOPEN_SOURCE
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
@@ -177,10 +179,12 @@ void close_proc()
 
 int set_proc_fd(int fd)
 {
-	return install_service_fd(PROC_FD_OFF, fd);
+	if (install_service_fd(PROC_FD_OFF, fd) < 0)
+		return -1;
+	return 0;
 }
 
-int set_proc_mountpoint(char *path)
+static int open_proc_sfd(char *path)
 {
 	int fd, ret;
 	close_proc();
@@ -212,7 +216,7 @@ inline int open_pid_proc(pid_t pid)
 
 	dfd = get_service_fd(PROC_FD_OFF);
 	if (dfd < 0) {
-		if (set_proc_mountpoint("/proc") < 0)
+		if (open_proc_sfd("/proc") < 0)
 			return -1;
 
 		dfd = get_service_fd(PROC_FD_OFF);
